/*
ComSpy is a program to listen on a com port, and dump what it hears in hex and ascii, like a classical dump
in a terminal window.  This is as simple as I could make it and still have it do what I wanted for debug
and reverse engineering uses.  Basically one uses a special cable to connect the data-in line of the port
ComSpy will use to a data line being used to communicate between two other devices.  If you need to see
both directions, you can start two instances of this attached to seperate ports.  Whatever you type on
the keyboard goes out too, except for Ctrl-c which will terminate the program.  To avoid complexity, this
is all in one file and declared in order so that no extra .h files are needed.  See usage for usage.

This was built and tested in RH8.

You have to su or add this to sudoers for it to be able to grab a com port in Linux.
 
I hope this program does something useful for you.  I thought about GPL'ing it, but decided nah, it's
not big or cool enough for that.  However, if you make any cool changes I'd appreciate hearing about it at
clab@starband.net.

See the projects web page at http://clab.mystarband.net
 
(C) Doug Coulter, no rights reserved
 
*/
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <ctype.h>
#include <stdio.h>
#include <string.h>  // String function definitions 
#include <fcntl.h>   // File control definitions 
#include <termios.h> // POSIX terminal control definitions
#include <time.h>    // for input timing functions
#include <pthread.h> // for threads
//#include <signal.h>   // temp for testing signals, take me out later
#include <sys/poll.h> // non blocking kbd read
#include <sys/time.h>  // for timestamps on incoming bytes

////////////////////////////////////////////////////////////
#define MAX_PATH 255         // logfilepath buffer size etc
#define DUMP_LINE_LEN 16     // traditional dump line length
/////////////////////////////////////
char InBuf[MAX_PATH+1]; // static global, rec buf
char OutBuf[MAX_PATH+1]; // printf buf
bool bRun = true;      // quit flag
int CharCount; // how many we actually got as input
pthread_t rThread; // pointer to input thread
char logfilepath[MAX_PATH+1]; // filepath to logfile, if used
bool buselogfile = false;
bool bTimeStamp = false; // use timestamps since startup
timeval curtime; // shared between read_port() and handle_input()
/////////////////////////////////////////
/*void sig_caught(int signum)
{ // perhaps we can do clean ctrl-c termination this way?
  printf("sig caught:%d\r\n",signum);
  bRun = false;
} */

/////////////////////////////////////////

// Test the linux serial port stuff


// 'open_port()' - Open serial port,
//  numbering starts with zero.
// Returns the file descriptor on success or -1 on error.


int open_port(int portno)
{
  int fd; // File descriptor for the port 
  char dev[20]; // more bytes than we should need (around 11 normally)
  char ermsg[50];
  sprintf(dev,"/dev/ttyS%d",portno);
  fd = open(dev, O_RDWR | O_NOCTTY | O_NDELAY);
  if (fd == -1)
  { // Could not open the port.
    strcpy(ermsg,"Could not open ");
    strcat(ermsg,dev);
    strcat(ermsg," - ");
    perror(ermsg);
    bRun = false; // may as well quit if we have no port
  } else fcntl(fd, F_SETFL, 0);
  return(fd);
}
/////////////////////////////////////////////////////////////////
int read_port(int pd)
{
  int retval,charcount;
  clock_t donetime;

  retval = read(pd,InBuf,DUMP_LINE_LEN); // try for some characters
  if (-1 == retval) return 0; // there was nothing to get

  if (bTimeStamp) // then get the current time for display
  {
    if (gettimeofday(&curtime,NULL))
    {
      bTimeStamp = false; // don't flood screen if something is broken
      perror("Couldn't get current time - ");
    }
  }
  if (DUMP_LINE_LEN == retval)  return retval; // we got totally lucky (or we are getting behind!).    
  // note the else is redundant...
  //else some partial count, try for awhile to get a full line, then return
  charcount = retval; // how many we've gotten so far
  donetime = clock(); // get current time in approximate uS
  donetime += CLOCKS_PER_SEC/2; // add time limit, hope for no rollover...

  while (charcount < DUMP_LINE_LEN && clock() < donetime) // add a time limit here
  {
    retval = read(pd,&InBuf[charcount],DUMP_LINE_LEN-charcount); // try for more
    if (-1 != retval) charcount += retval; // add in any new chars      
  }

  return charcount;
}

/////////////////////////////////////////////////////////////////
void* handle_input(void *fd)
{
  int pd = *(int *)fd;
  int i;
  const char hex_conv[17] = "0123456789ABCDEF"; 
  char tbuf[20];
  char obuf[150]; // we use 65 + timestamp + crlf nul
  char errbuf[MAX_PATH + 30]; // wasteful, kinda, but need no buffer overflow in error messages!
  int filed;
  timeval starttime;
  double printtime;

  if (buselogfile)
  {  // then open it, silly! FIXME -- needs some work on permissions

    filed = open(logfilepath,O_CREAT | O_TRUNC | O_WRONLY,0666); // let everyone play with this
    if (0 > filed)
    {
      errbuf[0] = 0; // make into a string
      strcat(errbuf, "Could not open:");
      strcat(errbuf,logfilepath);
      strcat(errbuf," - ");
      perror(errbuf); // nice pretty error message
      buselogfile = false; // don't cascade the errors, but run anyway
    }
  }

  if( bTimeStamp )
  {// get our startup time for later subtracts
   if (gettimeofday(&starttime,NULL)) // returns -1 for errors
   {
    perror("Couldn't get start time - ");
    bTimeStamp = false; // prevent error cascades
   }
  }
  ////////////////////// main receive loop
  while (bRun)
  {
    CharCount = read_port(pd);
    if (CharCount)
    {
      obuf[0] = 0; // null terminate

      for (i=0;i<CharCount;i++)
      {
        //	 sprintf(tbuf,"%2X ",InBuf[i]);
        tbuf[0] = hex_conv[(InBuf[i] >> 4) & 0xf]; // top nybble first
        tbuf[1] = hex_conv[InBuf[i] & 0xf];
        tbuf[2] = ' '; // space between hex bytes
        tbuf[3] = 0; // null terminate, doh!
        strcat(obuf,tbuf); // add to output buffer
      } // end for charcount, first part of dumpline

      while (strlen(obuf) < 48) strcat(obuf," "); // make a fixed line length, in a lame way

      for (i=0;i<CharCount;i++)
      {
        if (InBuf[i] > 0x1f && InBuf[i] < 0x7f) // then it's printable
        {
          sprintf(tbuf,"%c",InBuf[i]); // lazy dumb way to extract a char into a string
          strcat(obuf,tbuf); // add it in to output
        } else // print a period
        {
          strcat(obuf,".");
        }
      } // end for CharCount
      while (strlen(obuf) < 65) strcat(obuf," "); // make a fixed line length, in a lame way
      
      if(bTimeStamp)
      { // then format the diff between start and current time, and add to string at end
       printtime = curtime.tv_sec - starttime.tv_sec; // get the seconds part
       printtime += (double) (curtime.tv_usec - starttime.tv_usec)/1000000.; // get uS
       sprintf(tbuf,"%g sec",printtime);
       strcat(obuf,tbuf); // add this to display line
      }
      strcat(obuf,"\r\n");
      printf("%s",obuf); // do out business

      if (buselogfile)
      { // do the logfile stuff, as usual, error reporting is more work than the basic job
        if (0 > write(filed,obuf,strlen(obuf)))
        {
          errbuf[0] = 0;
          strcat(errbuf,"Error writing ");
          strcat(errbuf,logfilepath);
          strcat (errbuf," - ");
          perror(errbuf);
          close(filed); // close it anyway
          buselogfile = false; // stop trying any more on error
        } // if error
      } // if buselogfile
    } // if charcount    
  } // while brun
  close(filed); // done with this I suppose
  printf("Receive thread clean termination.\r\n");
  return NULL; // heh heh
}
/////////////////////////////////////////////////////////////////
void usage(void)
{ // instructions for ComSpy
  printf("Command line options begin with a minus sign, e.g. -h.\r\n");
  printf("Each option needs its own dash, they can't be combined\r\n");
  printf("Options are:\r\n");
  printf("-b or -B, followed by a legal number, sets baud rate.\r\n");
  printf("Legal baud rates are: 50,75,110,134,150,200,300,600,1200,\r\n");
  printf("1800,2400,4800,9600,19200,57600,115200,230400,460800,500000,\r\n");
  printf("576000,921600,1000000,1152000,1500000,2000000,2500000,3000000,\r\n");
  printf("3500000,4000000, but all may not work on your machine.\r\n\r\n");
  printf("-c or -C sets the com port.  The numbers begin at 1 (default).\r\n");
  printf("-h or -H turns on hardware handshaking, which is off otherwise.\r\n");
  printf("-l or -L sets char length.  5-8 are supported, default is 8\r\n");
  printf("-p or -P sets parity, options are N, E, O, N is default.\r\n");
  printf("-s or -S set stop bits, 1 or 2.  Default is 1.\r\n");
  printf("-t or -T turn on timestamping for each recieved group of bytes.\r\n");
  printf("<ctrl>-c terminates the program.\r\n");
  printf("-? or just ? gets this usage text.\r\n\r\n");
  printf("Any text not preceeded with a - is interpreted as a file path to\r\n");
  printf("accept logging data.  Options are not sensitive to the order they\r\n");
  printf("appear on the command line.  Default baudrate is currently 57600.\r\n");
}
/////////////////////////////////////////////////////////////////

int main (int argc, char **argv)
{
  int fd,n,i,itemp;
  struct termios options, orig_settings;
  struct termios term, orig_term;
  pollfd keybd;
  bool bHardwareHandshake = false;
  int ComPort = 0; // using linux numbering, comm 1
  speed_t baud = B57600; // unsigned int, a weird proliferation of types, set a default
  int bits = 8; // how many bits in serial port char
  int stops = 1; // how many stop bits
  bool parity = false; // defaults work out to 8n1
  bool oddparity = true; // same sense as actual bit
  char c;
 

  logfilepath[0] = 0; // make string null

/* // just playing with signals here, but decided to find another way to get the job done  
  if (signal(SIGINT,sig_caught) == SIG_ERR) printf("Failed to set sigint handler\r\n");
  if (signal(SIGQUIT,sig_caught) == SIG_ERR) printf("Failed to set sigquit handler\r\n");
  if (signal(SIGTERM,sig_caught) == SIG_ERR) printf("Failed to set sigterm handler\r\n"); */

  for (i=1;i<argc;i++)
  {
//        printf("arg#:%d:%s\r\n",i,argv[i]);
    if ('-' != *argv[i])
    { // strncpy used to prevent buffer overflow exploits
      if ('?' == *argv[i])
      {
        usage(); // most programs support this, so we will too
      } else
      {
        strncpy(logfilepath,argv[i],MAX_PATH); // only options with - aren't logfile paths, last one "wins"
        printf("Using logfile:%s\r\n",logfilepath);  // report this to user in case they didn't mean what they said
        buselogfile = true; // don't want to have to check the path length every time to see if we should write
      }
    } else
    { // we know this started out with a - sign
//         printf("test:%s\r\n",argv[i]+1);
      switch (*(argv[i]+1))
      {
      case 'b':
      case 'B': // baud rate
        itemp = atoi(argv[i]+2); // get it as a number we can switch on, gheesh, this bites rocks as an interface
        switch (itemp)
        { // this is needed as whoever designed termios used oddball bit patterns to define baudrates. Yech.
        case 50:     baud = B50; break; // pretty lame sparse table lookup
        case 75:     baud = B75; break;
        case 110:    baud = B110; break;
        case 134:    baud = B134; break;
        case 150:    baud = B150; break;
        case 200:    baud = B200; break;
        case 300:    baud = B300; break;
        case 600:    baud = B600; break;
        case 1200:   baud = B1200; break;
        case 1800:   baud = B1800; break;
        case 2400:   baud = B2400; break;
        case 4800:   baud = B4800; break;
        case 9600:   baud = B9600; break;
        case 19200:  baud = B19200; break;
        case 57600:  baud = B57600; break; // default at present, but default may change, so do it
        case 115200: baud = B115200; break;
        case 230400: baud = B230400; break; // this is about the most a normal (newer) PC will do, but ...
        case 460800: baud = B460800; break; // these are all in header file /usr/include/bits/termios.h too
        case 500000: baud = B500000; break;
        case 576000: baud = B576000; break;
        case 921600: baud = B921600; break;
        case 1000000:    baud = B1000000; break; // gheesh
        case 1152000:    baud = B1152000; break;
        case 1500000:    baud = B1500000; break;
        case 2000000:    baud = B2000000; break;
        case 2500000:    baud = B2500000; break;
        case 3000000:    baud = B3000000; break;
        case 3500000:    baud = B3500000; break;
        case 4000000:    baud = B4000000; break; // with speeds like this, who needs ethernet? ;)
        default:
          printf("%d is not a legal baudrate. Using default rate\r\n",itemp); //TODO I should be nicer and list them
          usage();
          break;
        }
        printf("baud rate:%d\r\n",itemp);
        break;

      case 'h': // hardware handshake on
      case 'H':
        bHardwareHandshake = true;
        break;
      case 'c': // comport # option
      case 'C':
        itemp = atoi(argv[i]+2); // get an integer
        if (itemp < 1)
        {
          printf("Comm ports are numbered 1 to howevermany. Can't open %d\r\n", itemp);
          printf("Will try comm port 1 \r\n");
          usage();
          itemp = 1;
        }
        itemp--; // linux starts with zero, humans start with one
        ComPort = itemp;               
        break;

      case 'l': // char length
      case 'L':
        bits = atoi(argv[i]+2); // get it as a number we can switch on
        if (bits < 5 || bits > 8)
        {
          printf("%d is not a valid value for bits/char. Using 8.\r\n",bits);
          bits = 8;
          usage();
        }
        break;
      case 'p':
      case 'P': // parity
        switch (*(argv[i]+2))
        {
        case 'n':
        case 'N':
          parity = false; // default, but defaults may change
          break;
        case 'o':
        case 'O':
          parity = true;
          oddparity = true; // default, but again, this may change
          break;
        case 'e':
        case 'E':
          parity = true;
          oddparity = false;
          break;
        default: // uh oh
          printf("%s is not a legal parity option.\r\n",argv[i]);
          usage();
        }
        break;
      case 's': // stop bits
      case 'S':
        stops = atoi(argv[i]+2); // get it as a number we can switch on
        if (stops < 1 || stops > 2)
        {
          printf("%d is not a legal number for stop bits, using 1.\r\n");
          stops = 1;
          usage();
        }
        break;
      case '?':
        usage();
        break;
      case 't': // timestamps
      case 'T':
        bTimeStamp = true; // flag for other thread
        break;
      default:
        printf("%s: is not an option ComSpy currently understands.\r\n",argv[i]);
        usage();
        break;

      } // end switch on minus sign options        
    } // if minus sign options        
  } // for() argc

  ////////////////////////////////////////////////// now set things up per user requests

// grab the port the user wanted, if possible    
  fd = open_port(ComPort); // linux port numbering starts with 0

// set up speeds and feeds

// Get the current options for the port...     
  tcgetattr(fd, &orig_settings);            // get a copy for later restore
  memcpy(&options,&orig_settings,sizeof(termios));

// Set the baud rates to user request or default above...
  cfsetispeed(&options, baud);
  cfsetospeed(&options, baud);

// Enable the receiver and set local mode...
  options.c_cflag |= (CLOCAL | CREAD);

// set parity options
  if (parity)
  {
    options.c_cflag |= PARENB;
    if (oddparity) options.c_cflag |= PARODD;
    else options.c_cflag &= ~PARODD;
  } else options.c_cflag &= ~PARENB;

// set stopbits
  if (1 == stops) options.c_cflag &= ~CSTOPB;
  else options.c_cflag |= CSTOPB;

// set bits per char   
  options.c_cflag &= ~CSIZE; // zero out current content
  switch (bits)
  {
  case 5:
    options.c_cflag |= CS5; // bring on the baudot!
    break;
  case 6:
    options.c_cflag |= CS6; // never seen this in 42 years of computers
    break;
  case 7:
    options.c_cflag |= CS7;
    break;
  case 8:
  default: // a belt for my suspenders, above
    options.c_cflag |= CS8; // default
    break;
  }

// hardware flow control
  if (bHardwareHandshake)
  {
    options.c_cflag |= CRTSCTS;
    printf("Hardware Handshake: ON\r\n");
  } else
  {
    options.c_cflag &= ~CRTSCTS;
    printf("Hardware Handshake: OFF\r\n");
  }

// set raw mode
  options.c_lflag &= ~(ICANON | ECHO | ECHOE | ISIG);

//  Set the new options for the port...
  tcsetattr(fd, TCSANOW, &options);
  fcntl(fd, F_SETFL, FNDELAY); // non blocking reads

  //////////////////////////////////////////////// now do the setup of the terminal end of things

  tcgetattr(fileno(stdin), &orig_term);            // get a copy for later restore
  memcpy(&term,&orig_term,sizeof(termios));       // mess with this copy
  term.c_lflag &= ~(ICANON | ECHO | ECHOE | ISIG | ECHOCTL); // turn off some things
  term.c_iflag &= ~(BRKINT | ICRNL);              // and yet some more
  tcsetattr(fileno(stdin),TCSANOW,&term);                 // do the deed
  /////////////////////////////////////////////// finally, run the program

  pthread_create(&rThread,NULL,handle_input,&fd); // receive/dump/log gets its own thread
                                                  // we then continue in this one and handle keyboard events
  keybd.fd = fileno(stdin);
  keybd.events = POLLIN;

  while (bRun)
  {
//     printf("loop\r\n");
    if (poll(&keybd,1,100) && keybd.revents)
    {

//     printf("keyevent:%x\r\n",keybd.revents); // for debug

      i = read(fileno(stdin),&c,1);
      if (i>0)
      {
        if ((c & 255) == 3) bRun = false;// ctrl-c is stop till we figure out something better

//      printf("key:%c, ordinal:%x\r\n",c,c); // debug
        n = write(fd,&c,1);
        if (n<0)
        {
          perror("Error writing to serial port:");
          bRun = false; // stop the music
        }
      } else if (i < 0)
      {// some sort of read error
        perror("Error reading stdin:");
        bRun = false;
      }
    }
  } // while bRun

// cleanup and leave

  pthread_join(rThread,NULL); // wait for other thread to exit before destroying process
  tcsetattr(fd,TCSANOW,&orig_settings);// should set it back the way it was first.
  close(fd); // close serial port
  tcsetattr(fileno(stdin),TCSANOW,&orig_term); // put things back for terminal window too
  printf("Send thread clean termination\r\n");
  return(0);
}

//@@@fixme it seems we don't always get the bash terminal back to normal on exit
